// Take a look at the license at the top of the repository in the LICENSE file.

pub(crate) struct ProductInner;

impl ProductInner {
    pub(crate) fn family() -> Option<String> {
        None
    }

    pub(crate) fn name() -> Option<String> {
        None
    }

    pub(crate) fn serial_number() -> Option<String> {
        None
    }

    pub(crate) fn stock_keeping_unit() -> Option<String> {
        None
    }

    pub(crate) fn uuid() -> Option<String> {
        None
    }

    pub(crate) fn version() -> Option<String> {
        None
    }

    pub(crate) fn vendor_name() -> Option<String> {
        None
    }
}
